/* 
   uFS Project [ufs.c]
   (c) Simon Kueppers
   20.12.06 Germany

   This file is the heart of the uFS. It provides basic page allocation/freeing functions.
   And also functions to parse pathstrings to adresses.
   This functions should only be used by ufs_dirio.c and ufs_fileio.c.
  
   For function descriptions look at ufs.h
   

   PS:
   I tried to comment the code as best as i was able to. This is a little bit confusing
   for a german, so please excuse if something is not that right it should be.
   Have fun with this code but remember the BEER LICENSE when you copy+paste something
   from here :-)

*/

#include <string.h>									//strcmp()
#include "stdint.h"									//Vartypedefs
#include "ufs.h"									//function prototypes and defines
#include "sd.h"										//function prototypes to access sd card

uint32_t pLastPage = 0;

/*
   Marks the page at 'adress' as free-to-use.
   Note: this function does not clear the page.

*/
void ufs_pfree(uint32_t adress)
{
	struct t_page TempPage;
	uint32_t temp;

	sd_read(pLastPage, &TempPage, 512);				//read last-page to buffer
	temp = TempPage.hdr_pNextpage;					//remember follower of last-page
	TempPage.hdr_pNextpage = adress;				//new follower of last-page is the free'd page
	sd_write(pLastPage, &TempPage);					//write last-page back without changing its content

	pLastPage = temp;								//set last-page to follower of free'd page

	sd_read(adress, &TempPage, 512);				//read free'd page to buffer
	TempPage.hdr_pNextpage = temp;					//set new follower to old last-page follower
	sd_write(adress, &TempPage);					//and write back without changing its content
}

/*
   Searches an empty page, marks it as 'used' and returns the adress.
   Note: the page allocation/freeing functions use a bitmappage buffer
   in order to make "burst-allocating" or "burst-freeing" faster.
   This should increase speed when deleting or creating large files.
*/
uint32_t ufs_palloc(void)
{
	struct t_page TempPage;
	uint32_t freepageadress, temp;

	sd_read(pLastPage, &freepageadress, 4);			//checkout adress of free page from last-page

	if ((pLastPage == 0) && (freepageadress == 0))	//check if adress of free page really points to a free page and not to itself
		return 0;

	sd_read(freepageadress, &temp, 4);				//checkout adress of follower of free page

	sd_read(pLastPage, &TempPage, 512);				//and write it back to last-page without changing its content
	TempPage.hdr_pNextpage = temp;					//now, the page with adress 'freepageadress' is not in single linked list
	sd_write(pLastPage, &TempPage);					//anymore.

	pLastPage = temp;								//Set last-page to follower of the allocated page. used to prevent wearout of memory

	return freepageadress;							//return 0 to indicate that there is no space left
}


/*
   This function adds a dir entry pointed by 'linkadress'
   in the directory pointed by 'diradress'.
*/
uint8_t ufs_addentry(uint32_t diradress, uint32_t linkadress)
{
	struct t_page TempPage;
	uint32_t currentpage;
	uint8_t i;

	sd_read(diradress, &TempPage, 512);					//Fetch header of current dir

	while((currentpage = TempPage.hdr_pNextpage) != 0)//Loop while not last page in chain
	{
		sd_read(currentpage, &TempPage, 512);			//if there is a following page in chain, read it
		
		if(TempPage.hdr_nValidbytes <= 500)			//Is there any place left to place a new dir entry?
		{
			for(i=0; i<N_DIRENTRYS; i++)			//Loop number-of-entrys-times
			{
				if(TempPage.dirpage.entry[i] == 0)	//Find an entry, that is unused (=0)
				{
					TempPage.dirpage.entry[i] = linkadress;	//Write the headeradress into dir-entrys
					TempPage.hdr_nValidbytes += 4;	//Update the number of valid bytes
					
					sd_write(currentpage, &TempPage);//Write page back to disk

					return 1;						//return 1
				}
			}
		}
	}

	//The program only gets here, if there was no space left to place the dir-entry

	currentpage = ufs_palloc();					//create a new page to hold dir entry

	if (!currentpage)								//check if enough space for another dir entry page
		return 0;									//and return 0

	TempPage.hdr_pNextpage = currentpage;			//write the adress of new page in the last page in chain
	sd_write(diradress, &TempPage);					//and write this updated page back to sd card
	
	for(i=0; i<N_DIRENTRYS; i++)
		TempPage.dirpage.entry[i] = 0;				//set all dir-entrys to zero
		
	TempPage.hdr_pNextpage = 0;						//No next page here
	TempPage.hdr_nValidbytes = 4;					//4 valid bytes in new page (equals one dir entry)

	TempPage.dirpage.entry[0] = linkadress;			//place dir entry at place-0

	sd_write(currentpage, &TempPage);				//and write this new page to sd card

	return 1;
}

/*
   This function searches the dir-entry in the parent dir
   of the file/dir pointed by 'linkadress' and deletes it.
   Note: in fact, this function frees a page, that does not
   contain dir-entrys anymore, BUT it does not (yet) fix
   a possible fragmenting, where it can be possible, that
   there are a few dir-entry-pages in chain, where each
   holds only one dir-link at one time.
*/
uint8_t ufs_delentry(uint32_t linkadress)
{
	struct t_page TempPage;
	uint32_t currentpage, prevadress, temp;
	uint8_t i;

	if (linkadress == 0)							//linkadress is not valid
		return 0;

	sd_read(linkadress, &TempPage, 512);					//fetch the headerpage at linkadress

/*Delete dir entry in parent dir
*/
	
	prevadress = TempPage.headpage.pParent;			//extract the parent-dir adress

	sd_read(TempPage.headpage.pParent, &TempPage, 512);	//and fetch page at this adress

	while((currentpage =TempPage.hdr_pNextpage) != 0)//Loop while not last page in chain
	{
		sd_read(currentpage, &TempPage, 512);			//if there is a following page in chain, read it

		for(i=0; i<N_DIRENTRYS; i++)				//Loop through each dir entry
		{
			if(TempPage.dirpage.entry[i] == linkadress)//Find suitable entry
			{
				TempPage.dirpage.entry[i] = 0;		//Delete link
				
				if(TempPage.hdr_nValidbytes >= 8)	//Are more than 8 bytes used in this page?
				{
					TempPage.hdr_nValidbytes -= 4;	//Update the number of valid bytes
					sd_write(currentpage, &TempPage);//Write page back to disk
					
				}else								//this is the last dir entry in page
				{	
					temp = TempPage.hdr_pNextpage;	//save adress of the pNextpage pointer

					sd_read(prevadress, &TempPage, 512);	//fetch page before the page, which is now empty
					TempPage.hdr_pNextpage = temp;	//put saved adress instead the adress of the empty page
					sd_write(prevadress, &TempPage);//update page

					ufs_pfree(currentpage);		//and free the now unused and unlinked page
				}


				return 1;							//return successful
			
			}
		}

		prevadress = currentpage;					//current page will be previous page furthermore
	}

	//when the program gets here, there must be an undefined error, because
	//every object (file/dir) MUST be linked somewhere in the parent dir.
	//if its not, then the used space of the file/dir is dead and can not be used anymore.

	return 0;
}

/*
   This function searches all dir-entrys of the directory pointed by
   'diradress' after a directory called 'name'
*/
uint32_t ufs_exists(uint32_t diradress, uint8_t* name)
{
	uint16_t i;										
	struct t_page TempPage;							
	uint32_t currentpage;
	uint32_t currentdirentry;

	if (diradress == 0)
		return 0;

	sd_read(diradress, &TempPage, 512);					//Fetch Headerpage of the directory to search in

	if (strcmp(name, "..") == 0)					//If ".."
	{
		if(TempPage.headpage.pParent)				//and if there is a parent dir
			return TempPage.headpage.pParent;		//return parent dir adress
		else
			return 0;								//there is no parent dir. return 0
	}

	while((currentpage = TempPage.hdr_pNextpage) != 0)//Loop while not last page in chain
	{
		sd_read(currentpage, &TempPage, 512);			//load next page to buffer if there is one
		
		for(i=0; i<N_DIRENTRYS; i++)				//Check each dir entry
		{
			if ((currentdirentry = TempPage.dirpage.entry[i]) != 0)	//Is pointer valid?
			{
				sd_read(currentdirentry, &TempPage, 512);//Fetch corresponding header page

				if (!strcmp(name, TempPage.headpage.objname))//compare name of object with searched name
				{
					return currentdirentry;			//If found, return adress
				}
				
				sd_read(currentpage, &TempPage, 512);	//else load dir-entry page back to buffer
				
			}
		}
	}
	
	return 0;										//Return 0 if not found

}


/*
   This function checks the headerpage at 'objadress' for
   the ISDIR attribute
*/
uint8_t ufs_isdir(uint32_t objadress)
{
	struct t_page TempPage;

	if (objadress == 0)								//check if objadress is valid
		return 0;
	
	sd_read(objadress, &TempPage, 512);					//read page

	return (TempPage.headpage.attrflags & ATTR_ISDIR);//and return ISDIR-bit
}

/*
   This function parses a whole path 'path' and returns the adress
   of the parsed directory/file
   Note: There is something like a "convention" about path'ing files and dirs.
   Filepathes can be something like that:
   /path/to/file.ext -> this path is relative to root ("/")
   path/to/file.ext  -> this path is relative to 'startadress'
   Dirpathes can additionally be something like this:
   /path/to/dir/     -> this path is relative to root ("/")
   path/to/dir/		 -> this path is relative to 'startadress'
*/
uint32_t ufs_getadress(uint32_t startadress, uint8_t *path)
{
	uint32_t objptr = startadress;
	uint8_t objname[255];							//Dir/Filename max. 254 chars
	uint8_t* pobjname = objname;
	uint8_t i=0;
	
	if (*path == '/')								//if path begins with '/'
	{
		objptr = ROOTDIR_ADR;						//switch to root-dir before continueing
		path++;										//go to next char
	}
	
	while(*path)									//while pathname has not ended
	{
		*pobjname = *path;							//copy one char from path to objname string

		if (*path == '/')							//if copied char was '/'
		{
			*pobjname = '\0';						//replace this '/' with an '\0'

			if (!((objname[0] == '/') || (objname[0] == '\0')))
				objptr = ufs_exists(objptr, objname);//checkout adress of object

			if (objptr == 0)						//if any object was not found, return error
				return 0;

			pobjname = objname;						//reset pointer
		}else
			pobjname++;								//if char was not '/' set objname-pointer to next char

		path++;										//set path-pointer to next char
	}
	
	*pobjname = '\0';								//at last put an string terminate to objname's end

	if (!((objname[0] == '/') || (objname[0] == '\0')))
		objptr = ufs_exists(objptr, objname);		//and parse the last object

	return objptr;									//then return adress of found object
}

/*
   This function parses a path 'path', checks if it is a dir
   and when it exists, it returns the adress of it.
*/
uint32_t ufs_getdiradress(uint32_t startadress, uint8_t* path)
{
	uint32_t ptemp;

	ptemp = ufs_getadress(startadress, path);		//get adress of object

	if (ufs_isdir(ptemp))							//check if it is a dir
		return ptemp;
	else 
		return 0;

}

/*
   This function parses a path 'path', checks if it is a file
   and when it exists, it returns the adress of it.
*/
uint32_t ufs_getfileadress(uint32_t startadress, uint8_t* path)
{
	uint32_t ptemp;

	ptemp = ufs_getadress(startadress, path);		//get adress of object

	if (ufs_isdir(ptemp))							//check if file/dir
		return 0;
	else 
		return ptemp;

}
